/*
 The search for section 6.2 of https://www.mathi.uni-heidelberg.de/~yves/Papers/radical.html
 
 Please also look at the ReadMe:
 https://www.mathi.uni-heidelberg.de/~yves/Papers/radical_data/aux/ReadMe.pdf
 */

#ifndef _DHO_T_h
#define _DHO_T_h

#include <string>
#include <iostream>
#include <array>

#include "vs2.h"

// a class for binary DHOs
template<size_t rank, size_t dimU=rank*(rank+1)/2>
class DHO_T{
	
public:
	static const size_t r {rank};
	static const size_t n {dimU};  // dimension of the mbient space.
	static const size_t n_r=(1<<r); // # elements of the DHO
	
	typedef vec2T<n> U_T;		// vectors
	typedef vs2T<U_T> US_T;		// vector spaces
	
	std::array<std::array<U_T,n_r> ,n_r> sch; //  dho[i] meet dho[j]
	US_T P; // Radikal
	US_T U; // Ambient Space
	typedef std::array<US_T,n_r> dhob_T; // elemente of the DHO
	typedef std::bitset<(1<<n)>  inDHO_T; // inzidence vektor of points of the DHO
private:
	dhob_T dhob;
	
public:
	DHO_T()=default;
	~DHO_T()=default;
	DHO_T(const DHO_T&)=default;
	DHO_T(DHO_T&&)=default;
	constexpr DHO_T& operator=(const DHO_T&)=default;
	constexpr DHO_T& operator=(DHO_T&&)=default;
	
	
	
	friend std::ostream& operator<<(std::ostream& s,const DHO_T& D)
	{
		for (size_t i=0; i<n_r; i++)
			s<<D.dhob[i];
		return s;
	}
	
	friend std::istream& operator>>(std::istream& s, DHO_T& D)
	{
		U_T v;
		for (size_t i=0; i<n_r; i++)
			for (size_t j=0; j<r; j++)
			{
				s>>v;
				D.dhob[i].add(v);
			}
		return s;
	}
	
	
	
	US_T& operator[](const size_t i)
	{
		return dhob[i];
	}
	
	US_T operator[](const size_t i) const
	{
		return dhob[i];
	}
	
	bool init_meet()
	{
		for (size_t i=0; i<n_r; i++)
			sch[i][i]=U_T::zero;
		for (size_t i=0; i<n_r; i++)
			for (size_t j=i+1; j<n_r; j++)
			{
				US_T s=meet(dhob[i],dhob[j]);
				if (s.dim()!=1) {
					std::cerr<<"dimension DHO["<<i<<"] meet DHO["<<j<<"] != 1 ("<<s.dim()<<") "<<std::endl;
					std::cout<<s<<"\n"<<std::endl;
					std::cout<<dhob[i]<<"\n"<<std::endl;
					std::cout<<dhob[j]<<std::endl;
					
					return false; //exit(1);
				}
				sch[i][j]=s.pkt(1);
				sch[j][i]=sch[i][j];
			}
		return true;
	}
	
	void init_P()
	{
		for (size_t i=0; i<n_r; i++)
			for (size_t j=i+1; j<n_r; j++)
				for (size_t k=i+1; k<n_r; k++)
					P.add(sch[i][j]+sch[j][k]+sch[k][i]);
		//std::cout<<" dim P = "<< P.dim()<<std::endl;
	}
	
	void init_U()
	{
		for (size_t e=0; e<n_r; e++)
		{
			for (size_t j=0; j<r; j++)
				U.add(dhob[e].base(j));
		}
		// std::cout<<" dim U = "<< U.dim()<<std::endl;
	}
	
	
	static std::array<size_t,n_r> invers(const std::array<size_t,n_r>& p)
	{
		std::array<size_t,n_r> r;
		for (size_t i=0;i<n_r;i++)
			r[p[i]]=i;
		return r;
	}
	
	bool is_type_H() const
	{
		US_T W=dhob[0];
		std::array<size_t,n_r> p;
		for (size_t s=0; s<n_r; s++)
			p[s]= W.index(sch[0][s]);
		p=invers(p);

		for (size_t s=0; s<n_r; s++)
			for (size_t t=0; t<n_r; t++)
				for (size_t u=0; u<n_r; u++)
					if (sch[p[s]][p[t]]+sch[p[s]][p[u]]!=sch[p[s]][p[s^t^u]]) {
						return false;
					}
		//std::cout<<"Is Typ (H) "<<std::endl;
		return true;
	}
	
	
	static size_t xi(const size_t x,const size_t e)
	{
		if (x==0) {
			return 0;
		}
		else if (x==e) {
			return 0;
		}
		return e;
	}
	
	
	static U_T xi(const U_T x,const U_T e)
	{
		if (x==U_T::zero) {
			return U_T::zero;
		}
		else if (x==e) {
			return U_T::zero;
		}
		return e;
	}
	
	bool is_type_D(const std::array<size_t,n_r>& p, const size_t e) const
	{
		for (size_t s=0; s<n_r; s++)
			for (size_t t=0; t<n_r; t++)
				for (size_t u=0; u<n_r; u++)
					if (sch[p[s]][p[t]]+sch[p[s]][p[u]]!=sch[p[s]][p[s^t^u^xi(s^t,e)^xi(t^u,e)^xi(u^s,e)]]) {
						return false;
					}
		//std::cout<<"Ist Typ (D) mit e= "<<e;
		//std::cout<<"  und p= "<<p<<std::endl;
		return true;
	}
	
	bool is_type_D() const
	{
		if(P.dim() == n-r+1)
		{
			US_T W=dhob[0];
			US_T E=meet(P,W);
			U_T e=E.pkt(1);
			
			std::array<size_t,n_r> p;
			for (size_t s=0; s<n_r; s++)
				p[s]= W.index(sch[0][s] + xi(sch[0][s]+e,e)); // Indizierung X(w)= dhob[s] mit w= p(s)

			return is_type_D(invers(p),W.index(e));
		}
		return false;
	}
	
	void reset()
	{
		for (size_t i=0; i<n_r; i++)
			dhob[i].reset();
		P.reset();
		U.reset();
	}
	
	void resetPU()
	{
		P.reset();
		U.reset();
	}
	
private:
	bool DHO_test_0()const
	{ // dim(S_e)=r
		for (size_t i=0; i<n_r; i++)
		{
			if (dhob[i].dim()!=r) {
				std::cerr<<"Dim DHO["<<i<<"] != "<<r<<" ("<<dhob[i].dim()<<") "<<std::endl;
				return false;
			}
		}
		return true;
	}
	
	bool DHO_test_1()
	{// intersection of two DHO elements is 1-dim
		return init_meet();
	}
	
	bool DHO_test_2()const
	{	// intersection of three DHO elements is empty
		// init_meet() must already have been called.
		for (size_t i=0; i<n_r; i++)
			for (size_t j=i+1; j<n_r; j++)
				for (size_t k=j+1; k<n_r; k++)
					if(sch[j][i]==sch[i][k])
					{
						std::cerr<<"DHO["<<j<<"] meet  DHO["<<i<<"] =  DHO["<<i<<"] meet DHO["<<k<<"]"<<" ("<<sch[j][i]<<") "<<std::endl;
						return false;
					}
		return true;
	}
	
	
public:
	
	
	bool DHO_test()
	{
		return DHO_test_0()&& DHO_test_1()&& DHO_test_2();
	}
	
	inDHO_T get_inDHO(){
		// generates incidence vector of the points of DHO
		inDHO_T inD;
		inD.reset();
		inD.set(0);
		
		for (size_t i=0; i<n_r; i++)
			for (size_t j=1; j<n_r; j++)
				inD.set(size_t(dhob[i].pkt(j)));
		return inD;
	}
	
	void init()
	{
		init_meet();
		DHO_test_2();
		init_U();
		init_P();
	}
	
};

#endif // _DHO_T_h

