/*
 The search for section 6.2 of https://www.mathi.uni-heidelberg.de/~yves/Papers/radical.html
 
 Please also look at the ReadMe:
 https://www.mathi.uni-heidelberg.de/~yves/Papers/radical_data/aux/ReadMe.pdf
 */

#ifndef __vec2T_h
#define __vec2T_h

#include <iostream>
#include <bitset>

// A class for binary vectors in the dimUv-dimensional space
template<size_t dimUv>
class vec2T{
	static_assert(dimUv<=64,"dimUv>64");
	typedef std::bitset<dimUv> l_dataT;
	l_dataT x;
public:
	static const size_t dimU {dimUv};
	constexpr static const vec2T zero {0};
	typedef size_t bs_int;
	
	vec2T()=default;
	~vec2T()=default;
	constexpr vec2T(const vec2T&)=default;
	constexpr vec2T(vec2T&&)=default;
	constexpr vec2T& operator=(const vec2T&)=default;
	constexpr vec2T& operator=(vec2T&&)=default;
	
	constexpr explicit vec2T(const l_dataT& a):x(a){};
	constexpr explicit operator bs_int() const { return x.to_ulong(); }
	
	constexpr vec2T operator+=(const vec2T& b)
	{
		x^=b.x;
		return (*this);
	}
	constexpr friend vec2T operator+(const vec2T& a,const vec2T& b){
		vec2T r {a};
		r+=b;
		return r;
	}
	
	constexpr vec2T operator-=(const vec2T& b)
	{
		return (*this)+=(b);
	}
	constexpr friend vec2T operator-(const vec2T& a,const vec2T& b){
		return a+b;
	}
	
	constexpr  bool operator[](const size_t i) const
	{
		return x[i];
	}
	
	constexpr friend bool operator==(const vec2T& a,const vec2T& b)
	{
		return a.x==b.x;
	}
	
	constexpr friend bool operator!=(const vec2T& a,const vec2T& b)
	{
		return !(a==b);
	}
	
	friend std::ostream& operator<<(std::ostream& s,const vec2T& x)
	{
		s<< x.x;
		return s;
	}
	
	friend std::istream& operator>>(std::istream& s, vec2T& x)
	{
		std::string ist;
		s>> ist;
		x.x = l_dataT(ist);
		return s;
	}
	
	template<size_t d>
	static constexpr vec2T  embed(const vec2T<d> v, const size_t offs=0)
	{// embed an vector v of an smaller vector sapce in the current vector.
		vec2T r {bs_int(v)};
		r.x<<=offs;
		return r;
	}
	
	static constexpr vec2T e(const size_t i) noexcept // e_0 ... e_{dimU-1}. The standard basis vectors
	{
		return vec2T(bs_int(1)<<i);
	}
	
	
	constexpr bool is_zero() const noexcept
	{
		return x.none();
	}
	
	constexpr size_t first() const noexcept // the coordinate of the first nonzero entry of a vector.
	{
		for (size_t i=0;i<dimU;i++)
			if (x.test(i))
				return i;
		
		return dimU; // Fehler x =0,
	}
	
};

#endif // __vec2T_h

